{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}
{-# OPTIONS -fno-warn-missing-methods #-}

module Vec3
        ( Vec3
        , magnitudeV3
        , normaliseV3
        , mulsV3
        , dotV3
        , clampV3
        , clipV3)
where

type Vec3
        = (Float, Float, Float)


instance Num Vec3 where
 (+) (x1, x2, x3) (y1, y2, y3)
        = (x1 + y1, x2 + y2, x3 + y3)

 (-) (x1, x2, x3) (y1, y2, y3)
        = (x1 - y1, x2 - y2, x3 - y3)

 (*) (x1, x2, x3) (y1, y2, y3)
        = (x1 * y1, x2 * y2, x3 * y3)


-- | Yield the magnitude of a vector.
magnitudeV3 :: Vec3 -> Float
magnitudeV3 (x, y, z)
        = sqrt (x * x + y * y + z * z)


-- | Normalise a vector to have unit length.
normaliseV3 :: Vec3 -> Vec3
normaliseV3 v
        = v `mulsV3` (1.0 / magnitudeV3 v) 


-- | Multiply a vector by a scalar.
mulsV3 :: Vec3 -> Float -> Vec3 
mulsV3 (x1, y1, z1) s
        = (s * x1, s * y1, s * z1)


-- | Compute the dot product of two vectors.
dotV3 :: Vec3 -> Vec3 -> Float
dotV3 (x1, y1, z1) (x2, y2, z2)
        = x1 * x2 + y1 * y2 + z1 * z2


-- | Clamp a vectors components to some minimum and maximum values.
clampV3 :: Vec3 -> Float -> Float -> Vec3
clampV3 (r, g, b) minVal maxVal
 = (clamp r, clamp g, clamp b)
 where
       clamp x 
        | x <= minVal   = minVal
        | x >= maxVal   = maxVal
        | otherwise     = x


-- | Clip a vector's components to some maxiumum value.
clipV3 :: Vec3 -> Float -> Vec3
clipV3 (r, g, b) maxVal
 = (clip r, clip g, clip b)
 where 
       clip x
        | x > maxVal    = maxVal
        | otherwise     = x
